# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

from pathlib import Path

from Qt import QtWidgets, QtCore

from ftrack_qt.widgets.dialogs import FileDialog


class FileBrowser(QtWidgets.QWidget):
    '''Browse Widget is a line edit with a browse button'''

    path_changed = QtCore.Signal(object)

    def __init__(self, parent=None):
        '''
        Initialize Browse widget
        '''
        super(FileBrowser, self).__init__(parent=parent)

        self._path_le = None
        self._browse_btn = None

        self.pre_build()
        self.build()
        self.post_build()

    def pre_build(self):
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(15, 0, 0, 0)
        self.layout().setSpacing(0)

    def build(self):
        '''Build widgets'''
        self._path_le = QtWidgets.QLineEdit()

        self.layout().addWidget(self._path_le, 20)

        self._browse_btn = QtWidgets.QPushButton('BROWSE')
        self._browse_btn.setObjectName('borderless')

        self.layout().addWidget(self._browse_btn)

    def post_build(self):
        '''Connect widget signals'''
        self._browse_btn.clicked.connect(self._browse_button_clicked)

    def get_path(self):
        '''Get path from the line edit'''
        return self._path_le.text()

    def set_path(self, path_text):
        '''Set path to the line edit'''
        self._path_le.setText(path_text)

    def set_tool_tip(self, tooltip_text):
        '''Set tooltip'''
        self._path_le.setToolTip(tooltip_text)

    def _browse_button_clicked(self):
        '''Browse button clicked signal'''
        start_dir = self._path_le.text() or Path.home()
        choosen_file_path = FileDialog(
            start_dir=str(start_dir), dialog_filter="*"
        )
        self.set_path(choosen_file_path.path)
        self.path_changed.emit(choosen_file_path.path)



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack
# TODO: Clean this code
import time
import os
import logging
import urllib.request, urllib.parse, urllib.error

from Qt import QtCore, QtGui, QtWidgets
import shiboken2

from ftrack_utils.threading import BaseThread

# Cache of thumbnail images.
IMAGE_CACHE = dict()


class SessionThumbnailBase(QtWidgets.QLabel):
    '''Widget to load thumbnails from ftrack server.'''

    MAX_CONNECTIONS = 10  # Maximum number of parallel connections to allow

    thumbnailFetched = QtCore.Signal(object)
    thumbnailNotFound = QtCore.Signal()

    _connection_count = 0

    def __init__(self, session, scale=True, parent=None):
        super(SessionThumbnailBase, self).__init__(parent)
        self.session = session
        self._alive = True
        self._scale = scale

        self.logger = logging.getLogger(
            __name__ + '.' + self.__class__.__name__
        )

        # self._worker = None
        self.__loadingReference = None
        self.pre_build()
        self.post_build()

    def pre_build(self):
        self.thumbnailCache = {}
        self.setFrameStyle(QtWidgets.QFrame.StyledPanel)
        self.setAlignment(QtCore.Qt.AlignCenter)

        self.placholderThumbnail = (
            self.session.server_url + '/img/thumbnail2.png'
        )

    def post_build(self):
        self.thumbnailFetched.connect(self._downloaded)
        self.thumbnailNotFound.connect(self.use_placeholder)

    def load(self, reference):
        '''Load thumbnail from *reference* and display it.'''

        if reference in IMAGE_CACHE:
            if IMAGE_CACHE[reference] is not None:
                self._updatePixmapData(IMAGE_CACHE[reference])
            else:
                self._updateWithPlaceholderPixmap()
            return

        thread = BaseThread(
            name='get_thumbnail_thread',
            target=self._download_async,
            callback=self._downloaded_async,
            target_args=[reference],
        )
        thread.start()

        self.__loadingReference = reference

    def _download_async(self, reference):
        '''(Run in background thread) Download image'''
        while (
            SessionThumbnailBase.MAX_CONNECTIONS
            <= SessionThumbnailBase._connection_count
        ):
            time.sleep(0.01)
            # Thumbnail widget still active?
            if not shiboken2.isValid(self):
                return
        SessionThumbnailBase._connection_count += 1
        try:
            return self._download(reference)
        except urllib.error.URLError:
            # Not found
            if not shiboken2.isValid(self):
                # Thumbnail widget has been destroyed
                return
            self.thumbnailNotFound.emit()
        finally:
            SessionThumbnailBase._connection_count -= 1

    def _downloaded_async(self, html):
        '''(Run in background thread) Image has been downloaded, propagate to QT thread'''
        if not shiboken2.isValid(self):
            # Thumbnail widget has been destroyed
            return
        self.thumbnailFetched.emit(html)

    def _downloaded(self, result):
        '''Handler worker finished event.'''
        if not shiboken2.isValid(self):
            # Thumbnail widget has been destroyed
            return
        IMAGE_CACHE[self.__loadingReference] = result
        self._updatePixmapData(result)

        self.__loadingReference = None

    def use_placeholder(self):
        '''Use placeholder image'''
        IMAGE_CACHE[self.__loadingReference] = None
        self._updateWithPlaceholderPixmap()

    def _updatePixmapData(self, data):
        '''Update thumbnail with *data*'''
        if data:
            pixmap = QtGui.QPixmap()
            pixmap.loadFromData(data)
            self._scaleAndSetPixmap(pixmap)

    def _updateWithPlaceholderPixmap(self):
        '''Update thumbnail with default placeholder image'''
        resource_path = ':ftrack/image/default/placeholderThumbnail'
        self._scaleAndSetPixmap(QtGui.QPixmap(resource_path))

    def _scaleAndSetPixmap(self, pixmap):
        '''Scale and set *pixmap*.'''
        if self._scale:
            scaled_pixmap = pixmap.scaledToWidth(
                self.width(), mode=QtCore.Qt.SmoothTransformation
            )
        else:
            scaled_pixmap = pixmap
        self.setPixmap(scaled_pixmap)

    def _safeDownload(self, url, opener_callback, timeout=5):
        '''Check *url* through the given *opener_callback*.

        .. note::

           A placeholder image will be returned if there is not response within
           the given *timeout*.

        '''

        return opener_callback(url, timeout=timeout)

    def _download(self, url):
        '''Return thumbnail file from *url*.'''
        if url:
            ftrackProxy = os.getenv('FTRACK_PROXY', '')
            ftrackServer = self.session._server_url

            if ftrackProxy != '':
                if ftrackServer.startswith('https'):
                    httpHandle = 'https'
                else:
                    httpHandle = 'http'

                proxy = urllib.request.ProxyHandler({httpHandle: ftrackProxy})
                opener = urllib.request.build_opener(proxy)
                response = self._safeDownload(url, opener.open)
                html = response.read()
            else:
                response = self._safeDownload(url, urllib.request.urlopen)
                html = response.read()

            return html

        self.logger.warning('There is no url image to download')
        return None



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack
import os

from Qt import QtWidgets, QtCore


class FileDialog(QtWidgets.QFileDialog):
    '''File dialog Widget'''

    caption = 'Choose file'

    @property
    def path(self):
        '''Return selected path in the file dialog'''
        return self._path

    def __init__(self, start_dir, dialog_filter, parent=None):
        '''
        Initialize File dialog
        '''
        super(FileDialog, self).__init__(parent=parent)
        self._path = None
        (
            file_path,
            unused_selected_filter,
        ) = self.getOpenFileName(
            caption=self.caption, dir=start_dir, filter=dialog_filter
        )

        if not file_path:
            return

        self.proces_path(file_path)

    def proces_path(self, file_path):
        '''Process returned path of the file dialog'''
        self._path = os.path.normpath(file_path)



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

from Qt import QtGui, QtCore, QtWidgets

from ftrack_qt.utils.widget import (
    get_main_window_from_widget,
    get_framework_main_dialog,
)

from ftrack_qt.widgets.icons import MaterialIcon


class OverlayWidget(QtWidgets.QFrame):
    '''
    Display a semi-transparent overlay over another widget.

    While the overlay is active, the main parent window and its children will not
    receive interaction events from the user (e.g. focus).
    '''

    def __init__(
        self,
        widget,
        width_percentage=0.95,
        height_percentage=0.6,
        transparent_background=False,
        parent=None,
    ):
        '''
        Initialize Overlay

        :param widget: The widget to put inside the overlay
        :param width_percentage: The percentage (0-100) of width it should consume
        :param height_percentage: The percentage (0-100) of height it should consume
        :param transparent_background: If True(default), overlay background should be made semi transparent
        :param parent: The parent dialog or frame
        '''
        super(OverlayWidget, self).__init__(parent=parent)

        self._widget = widget
        self._transparent_background = transparent_background
        self._width_percentage = width_percentage
        self._height_percentage = height_percentage

        self._container_widget = None
        self._close_btn = None
        self._fill_color = None
        self._pen_color = None
        self._event_filter_installed = False

        self.build()
        self.post_build()

    def build(self):
        # Create a container widget to hold the widget to be overlaid.
        self._container_widget = QtWidgets.QFrame(parent=self)
        self._container_widget.setLayout(QtWidgets.QVBoxLayout())

        h_layout = QtWidgets.QHBoxLayout()
        h_layout.setContentsMargins(0, 0, 0, 0)
        h_layout.addStretch()

        self._close_btn = QtWidgets.QPushButton(
            '', parent=self._container_widget
        )
        self._close_btn.setIcon(MaterialIcon('close', color='#D3D4D6'))
        self._close_btn.setObjectName('borderless')
        self._close_btn.setFixedSize(24, 24)

        h_layout.addWidget(self._close_btn)

        self._container_widget.layout().addLayout(h_layout)

        self._container_widget.layout().addWidget(self._widget)
        self._container_widget.setAutoFillBackground(False)
        if self._transparent_background:
            self._container_widget.setAutoFillBackground(False)
            self._container_widget.setStyleSheet('background: transparent;')
        else:
            self._container_widget.setAutoFillBackground(True)
            self._container_widget.setStyleSheet('background: #1A2027;')

        self._fill_color = QtGui.QColor(26, 32, 39, 150)
        self._pen_color = QtGui.QColor("#1A2027")

    def post_build(self):
        self._close_btn.clicked.connect(self.close)

    def __del__(self):
        if self._event_filter_installed:
            application = QtCore.QCoreApplication.instance()
            application.removeEventFilter(self)

    def paintEvent(self, event):
        '''(Override)'''
        super(OverlayWidget, self).paintEvent(event)
        # Paint a transparent overlay over the parent widget.
        size = self.size()  # get current window size
        painter = QtGui.QPainter()
        painter.begin(self)
        painter.setRenderHint(QtGui.QPainter.Antialiasing, True)
        painter.setPen(self._pen_color)
        painter.setBrush(self._fill_color)
        painter.drawRect(0, 0, size.width(), size.height())
        painter.end()

    def resizeEvent(self, event):
        '''(Override)'''
        super(OverlayWidget, self).resizeEvent(event)
        size = self.size()
        widget_width = size.width() * self._width_percentage
        widget_height = size.height() * self._height_percentage
        widget_x = int((size.width() - widget_width) / 2)
        widget_y = 40  # int(size.height()/2-widget_height/2)
        self._container_widget.resize(widget_width, widget_height)
        self._container_widget.move(widget_x, widget_y)

    def setVisible(self, visible):
        '''(Override) Set whether *visible* or not.'''
        main_window = get_framework_main_dialog(self._container_widget)
        if not main_window:
            main_window = get_main_window_from_widget(self)
        if visible:
            if not self._event_filter_installed:
                # Install global event filter that will deal with matching parent size
                # and disabling parent interaction when overlay is visible.
                main_window.installEventFilter(self)
                self._event_filter_installed = True
            # Manually clear focus from any widget that is overlaid. This
            # works in conjunction with :py:meth`eventFilter` to prevent
            # interaction with overlaid widgets.
            parent = self.parent()
            if parent.hasFocus():
                parent.clearFocus()
            else:
                for widget in parent.findChildren(QtWidgets.QWidget):
                    if self.isAncestorOf(widget):
                        # Ignore widgets that are part of the overlay.
                        continue

                    if widget.hasFocus():
                        widget.clearFocus()
                        break
        else:
            if self._event_filter_installed:
                main_window.removeEventFilter(self)
                self._event_filter_installed = False

        super(OverlayWidget, self).setVisible(visible)
        if visible:
            # Make sure size is correct
            if main_window:
                self.resize(main_window.size())

    def eventFilter(self, obj, event):
        '''Filter *event* sent to *obj*.

        Maintain sizing of this overlay to match parent widget.

        Disable parent widget of this overlay receiving interaction events
        while this overlay is active.

        '''
        if not isinstance(obj, QtCore.QObject) or not isinstance(
            event, QtCore.QEvent
        ):
            return False

        # Match sizing of parent.
        if obj == self.parent():
            if event.type() == QtCore.QEvent.Resize:
                # Relay event.
                self.resize(event.size())
        return False

        # Let event propagate.
        return super(OverlayWidget, self).eventFilter(obj, event)



from ftrack_qt.widgets.icons.material_icon import MaterialIcon
from ftrack_qt.widgets.icons.arrow_icon import ArrowMaterialIconWidget
from ftrack_qt.widgets.icons.status_icon import StatusMaterialIconWidget



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack
import logging

from Qt import QtCore, QtWidgets, QtGui, QtSvg


class MaterialIcon(QtGui.QIcon):
    '''Material icon, displaying SVG material icon images'''

    def __init__(self, name, color=None, variant=None, parent=None):
        '''
        Initialize the MaterialIcon

        :param name: The name of material icon SVG image
        :param color: The color, in html #RRGGBB format, or rgba(r,g,b,alpha)
        :param variant: The variant of material icon to use
        :param parent:  The parent dialog or frame
        '''

        self.logger = logging.getLogger(
            __name__ + '.' + self.__class__.__name__
        )

        self._name = name.replace('-', '_')
        self.color = color
        if variant is None:
            variant = 'filled'
        resource_path = ':ftrack/image/material-design-icons/{}/{}'.format(
            variant, self._name
        )
        pixmap = None
        if not color is None:
            # Read SVG and add fill color
            inFile = QtCore.QFile(resource_path)
            if inFile.open(QtCore.QFile.ReadOnly | QtCore.QFile.Text):
                text_stream = QtCore.QTextStream(inFile)
                svg_data = text_stream.readAll()
                svg_data = svg_data.replace(
                    '/></svg>', ' fill="{}"/></svg>'.format(color)
                )
                svg_renderer = QtSvg.QSvgRenderer(
                    QtCore.QByteArray(bytearray(svg_data.encode()))
                )
                pixmap = QtGui.QPixmap(svg_renderer.defaultSize())
                pixmap.fill(QtCore.Qt.transparent)
                painter = QtGui.QPainter(pixmap)
                svg_renderer.render(painter)
                painter.end()
            else:
                self.logger.warning(
                    'Unknown material icon resource: {}!'.format(resource_path)
                )
        else:
            pixmap = QtGui.QPixmap(resource_path)
            if pixmap is None or pixmap.isNull():
                self.logger.warning(
                    'Unknown material icon resource: {}!'.format(resource_path)
                )
        super(MaterialIcon, self).__init__(pixmap, parent=parent)



import logging

from Qt import QtWidgets

import ftrack_constants as constants

from ftrack_qt.utils.layout import recursive_clear_layout
from ftrack_qt.widgets.overlay import OverlayWidget
from ftrack_qt.widgets.buttons import (
    ProgressStatusButtonWidget,
    ProgressPhaseButtonWidget,
)


class ProgressWidget(QtWidgets.QWidget):
    '''
    Widget representation of the progress overlay widget used during execution of a
    lengthy proces that needs to visualize status updates and provide log
    feedback.

    The widget is composed of a dialog header docked button widget and info button
    widget representing main status. Main area is a  scroll area containing detailed
    progress feedback in form of phases (plugin runs) grouped into categories
    (context, collector...).

    '''

    MARGINS = 15

    @property
    def status_widget(self):
        '''Return the status widget'''
        return self._status_widget

    @property
    def status(self):
        '''Return the overall progress widget status'''
        return self.status_widget.status

    @property
    def action(self):
        '''Return a descriptive name of the action that is being run'''
        return self._action or ''

    def __init__(self, action, data, parent=None):
        '''Initialise ProgressWidgetObject with *action* describing what progress
        represents and *data* containing information about the progress phases,
        with ptional *parent*'''

        super(ProgressWidget, self).__init__(parent=parent)

        self._overlay_container = None
        self._content_widget = None
        self._status_banner = None
        self._status_widget = None
        self._scroll = None
        self._categories = []
        self._phase_widgets = {}
        self._main_window = None
        self._action = action

        self.logger = logging.getLogger(__name__)

        self.build()
        self.post_build()

        self.set_data(data)

    def build(self):
        self._status_widget = ProgressStatusButtonWidget('header-button')
        self._set_status_widget_visibility(False)

        self._scroll = QtWidgets.QScrollArea()
        self._scroll.setStyle(QtWidgets.QStyleFactory.create("plastique"))
        self._scroll.setWidgetResizable(True)

        self._content_widget = QtWidgets.QFrame()
        self._content_widget.setObjectName('overlay')
        self._content_widget.setLayout(QtWidgets.QVBoxLayout())
        self._content_widget.layout().setContentsMargins(
            self.MARGINS, self.MARGINS, self.MARGINS, self.MARGINS
        )

        self._scroll.setWidget(self._content_widget)

        self._overlay_container = OverlayWidget(
            self._scroll, height_percentage=0.8
        )
        self._overlay_container.setVisible(False)

    def post_build(self):
        '''Wire up signals'''
        self.status_widget.clicked.connect(self.show_overlay)

    def set_data(self, data):
        '''Set the data for the progress widget, were *data* is a list of
        dictionaries, one for each progress phase, with the following keys:
            - id: Unique id of the phase
            - label: A label for the phase
            - category: (optional) The category of the phase
            - tags: (optional) A list of tags for the phase
            - indent: (optional) The indent level of the phase
        '''
        assert data and isinstance(data, list), 'Data must be a list'

        recursive_clear_layout(self._content_widget.layout())

        self._status_banner = ProgressStatusButtonWidget('overlay-banner')
        self._content_widget.layout().addWidget(self._status_banner)

        self._categories = []
        self._phase_widgets = {}

        for phase_data in data:
            assert phase_data.get(
                'id'
            ), 'ID must be defined with progress widget phase!'
            assert phase_data.get(
                'label'
            ), 'Label must be defined with progress widget phase!'

            id_ = phase_data['id']
            label = phase_data['label']
            category = phase_data.get('category', '')
            tags = phase_data.get('tags')
            indent = phase_data.get('indent', 0)

            phase_button = ProgressPhaseButtonWidget(
                label, category=category, tags=tags
            )
            self._phase_widgets[id_] = phase_button
            self._content_widget.layout().setContentsMargins(
                self.MARGINS + indent * 10,
                self.MARGINS,
                self.MARGINS,
                self.MARGINS,
            )
            if category not in self._categories:
                self._categories.append(category)
                phase_category = QtWidgets.QLabel(category)
                phase_category.setObjectName("gray")
                self._content_widget.layout().addWidget(phase_category)
            self._content_widget.layout().addWidget(phase_button)

        self._content_widget.layout().addWidget(QtWidgets.QLabel(), 10)

    def _set_status_widget_visibility(self, visibility=False):
        '''Update the visibility of the progress widget'''
        self.status_widget.setVisible(visibility)

    def show_overlay(self, main_window=None):
        '''Show the progress widget overlay on top of *main_window*'''
        if main_window:
            self._main_window = main_window
            self._overlay_container.setParent(self._main_window)
        self._overlay_container.setVisible(True)
        self.status_widget.setVisible(True)

    def hide_overlay(self):
        '''Hide the progress widget'''
        self.status_widget.setVisible(False)

    # Run
    def reset_statuses(self, new_status=None, status_message=''):
        '''Reset statuses of all progress phases'''
        if not new_status:
            new_status = constants.status.UNKNOWN_STATUS
        for phase_widget in list(self._phase_widgets.values()):
            phase_widget.update_status(new_status, status_message, None)
            phase_widget.hide_log()

    def run(self, main_widget):
        '''Run progress widget on top of *main_widget*, with *action*'''
        self.reset_statuses()
        self.update_status(
            constants.status.RUNNING_STATUS,
            message=f'Running {self.action.lower()}...',
        )
        self.show_overlay(main_widget)

    def update_status(self, new_status, message=None):
        '''Set the new main status to *new_status*, with optional *message*'''
        self.logger.debug(
            f'Main status update: {new_status} (message: {message})'
        )
        self.status_widget.set_status(new_status, message=message)
        if self._status_banner:
            self._status_banner.set_status(new_status, message=message)
        self._set_status_widget_visibility(True)

    def update_phase_status(
        self,
        id_,
        new_status,
        log_message='',
        status_message=None,
        time=None,
    ):
        '''Update the status of a phase/plugin *id_* to *new_status*, with
        optional *log_message*, *status_message* and execution *time*'''
        self.logger.debug(
            f'Phase {id_} status update: {new_status} (message: {log_message}'
        )
        assert id_, 'Widget ID cannot be None'
        assert new_status, 'Status cannot be None'
        if not status_message:
            status_message = (
                constants.status.STATUS_STRING_MAPPING.get(new_status)
                or new_status
            )

        if id_ in self._phase_widgets:
            phase_widget = self._phase_widgets[id_]
            phase_widget.update_status(
                new_status, status_message, log_message, time=time
            )
            main_status_message = (
                f'{phase_widget.category}.'
                f'{phase_widget.text()}: {status_message}'
            )
            self.update_status(self.status, message=main_status_message)

        else:
            self.logger.warning(
                f'Progress phase widget with ID {id_} not found!'
            )

        # Error or finished?
        if new_status in [
            constants.status.ERROR_STATUS,
            constants.status.EXCEPTION_STATUS,
        ]:
            # Failed, reflect on main status
            if new_status == constants.status.ERROR_STATUS:
                self.update_status(
                    constants.status.ERROR_STATUS,
                    message=f'{self.action.title()} failed, see logs for details.',
                )
            elif new_status == constants.status.ERROR_STATUS:
                self.update_status(
                    constants.status.EXCEPTION_STATUS,
                    message=f'{self.action.title()} CRASHED, see logs for details.',
                )
        elif new_status == constants.status.SUCCESS_STATUS:
            finished = True
            for ref, widget in list(self._phase_widgets.items()):
                if widget.status != constants.status.SUCCESS_STATUS:
                    finished = False
                    break
            if finished:
                self.update_status(
                    constants.status.SUCCESS_STATUS,
                    message=f'{self.action.title()} completed successfully',
                )

    def teardown(self):
        '''Teardown the progress widget - properly cleanup the overlay containers'''
        self.hide_overlay()
        for phase_widget in list(self._phase_widgets.values()):
            phase_widget.teardown()
        self._overlay_container.setVisible(False)
        self._overlay_container.deleteLater()



from ftrack_qt.widgets.thumbnails.asset_version import AssetVersionThumbnail
from ftrack_qt.widgets.thumbnails.base import ThumbnailBase
from ftrack_qt.widgets.thumbnails.context import ContextThumbnail
from ftrack_qt.widgets.thumbnails.ellipse import EllipseThumbnailBase
from ftrack_qt.widgets.thumbnails.user import UserThumbnail
from ftrack_qt.widgets.thumbnails.session_base import SessionThumbnailBase



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

import logging

from Qt import QtWidgets, QtCore, QtGui


class AssetList(QtWidgets.QListWidget):
    '''Widget presenting list of existing assets'''

    version_changed = QtCore.Signal(object)
    '''Signal emitted when version is changed, with assetversion entity as 
    argument (version select mode)'''

    assets_added = QtCore.Signal(object)
    '''Signal emitted when assets are added to the list'''

    selected_item_changed = QtCore.Signal(object, object, object)
    '''Signal emitted when selected item is changed with arguments containing 
    the index of the current version, the version dictionary and the asset_id'''

    @property
    def latest_published_asset_item(self):
        return self._latest_published_asset_item

    def __init__(self, asset_list_widget_item, parent=None):
        '''Initialize AssetList'''
        super(AssetList, self).__init__(parent=parent)
        self.logger = logging.getLogger(
            __name__ + '.' + self.__class__.__name__
        )
        self.asset_list_widget_item = asset_list_widget_item
        self._latest_published_asset_item = None

        self.setHorizontalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(QtCore.Qt.ScrollBarAlwaysOff)
        self.setSpacing(1)
        self.assets = []
        self.currentItemChanged.connect(self._on_item_changed)

    def set_assets(self, assets):
        '''Set assets for the list'''
        self.assets = assets

        for asset_id, asset_dict in self.assets.items():
            widget = self.asset_list_widget_item(
                asset_name=asset_dict['name'],
                asset_id=asset_id,
                versions=asset_dict['versions'],
                server_url=asset_dict['server_url'],
            )

            if hasattr(widget, 'version_changed'):
                widget.version_changed.connect(
                    self._on_version_changed_callback
                )
            list_item = QtWidgets.QListWidgetItem(self)
            list_item.setSizeHint(
                QtCore.QSize(
                    widget.sizeHint().width(), widget.sizeHint().height() + 5
                )
            )
            if hasattr(widget, 'enable_version_select'):
                widget.enable_version_select = False
            self.setItemWidget(list_item, widget)
            self.addItem(list_item)
            # Find the latest published asset version.
            if not self._latest_published_asset_item:
                self._latest_published_asset_item = list_item
            else:
                if (
                    widget.version['date']
                    > self.itemWidget(
                        self._latest_published_asset_item
                    ).version['date']
                ):
                    self._latest_published_asset_item = list_item
        self.assets_added.emit(assets)
        # Pre select the latest published asset version.
        self.setCurrentItem(self._latest_published_asset_item)

    def _on_version_changed_callback(self, version):
        '''Handle version changed callback'''
        self.version_changed.emit(version)

    def _on_item_changed(self, current_item, previous_item):
        '''Handle item changed event'''
        if hasattr(self.itemWidget(current_item), 'enable_version_select'):
            if previous_item:
                self.itemWidget(previous_item).enable_version_select = False
            self.itemWidget(current_item).enable_version_select = True
        self.selected_item_changed.emit(
            self.indexFromItem(current_item),
            self.itemWidget(current_item).version,
            self.itemWidget(current_item).asset_id,
        )



from ftrack_qt.widgets.frames.asset_version_creation import (
    AssetVersionCreation,
)
from ftrack_qt.widgets.frames.asset_version_selection import (
    AssetVersionSelection,
)
from ftrack_qt.widgets.frames.new_asset_input import NewAssetInput



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack


def invoke_in_qt_main_thread(func):
    '''Decorator to ensure the function runs in the QT main thread.'''
    import ftrack_qt.utils.threading

    def wrapper(*args, **kwargs):
        return ftrack_qt.utils.threading.invoke_in_qt_main_thread(
            func, *args, **kwargs
        )

    return wrapper



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

from Qt import QtWidgets, QtCore, QtGui

from ftrack_qt.widgets.overlay import ShadedWidget

import ftrack_constants.qt as qt_constants
from ftrack_qt.utils.theme import apply_theme


class StyledDialog(QtWidgets.QDialog):
    '''
    The base dialog -intended to live docked, on top of DCC app main window or os
    a modal dialog in general.

    To be inherited by publisher, assembler and other dialogs.

    Designed to become shaded when another (modal) dialog is put in front of it,
    for visibility.
    '''

    DEFAULT_STYLE = qt_constants.theme.DEFAULT_STYLE

    # Allow child classes to override the default theme and background style
    theme = qt_constants.theme.DEFAULT_THEME
    background_style = qt_constants.theme.DEFAULT_BACKGROUND_STYLE
    docked = False

    @property
    def darken(self):
        return self._darken

    @darken.setter
    def darken(self, value):
        self._darken = value
        if self._darken:
            self._shaded_widget = ShadedWidget(self)
            self._shaded_widget.move(0, 0)
            self._shaded_widget.resize(self.size())
            self._shaded_widget.show()
        else:
            if self._shaded_widget:
                self._shaded_widget.close()

    def __init__(self, background_style=None, docked=False, parent=None):
        super(StyledDialog, self).__init__(parent=parent)
        self._darken = False
        self._shaded_widget = None

        if background_style:
            self.background_style = background_style

        if docked:
            self.docked = docked

        # Apply theme and with DCC specific properties
        # TODO: This has been deactivated as it makes the filebrowser dialog
        #  crash after selecting a file. If this needs to be activated in the
        #  future for DCC reasons: 1- Double check file browser dialog still
        #  working after selecting a file in standalone ui mode. 2- if not
        #  working, find another way to simulate the tool property in PySide2.
        #  Tool property documentation:
        #  https://doc.qt.io/qtforpython-5/PySide2/QtCore/Qt.html#:~:text=Qt.Tool-,Indicates,-that%20the%20widget
        #  probably the reason of the crashing remains in here: " This means that
        #  the window lives on a level above normal windows making it impossible
        #  to put a normal window on top of it. By default, tool windows will
        #  disappear when the application is inactive. This can be controlled by
        #  the WA_MacAlwaysShowToolWindow attribute. "
        # self.setWindowFlags(QtCore.Qt.Tool)

        self.setSizeGripEnabled(True)  # Enable resize on Windows

        apply_theme(self, self.theme)
        self.setProperty('background', self.background_style)
        self.setProperty('docked', 'true' if self.docked else 'false')
        self.setAttribute(QtCore.Qt.WA_DeleteOnClose, True)

        # Make sure the dialog is always on top
        self.setWindowFlags(QtCore.Qt.WindowStaysOnTopHint)

        # Have a proper title instead of default 'python'
        self.setWindowTitle('ftrack')



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

from Qt import QtWidgets, QtCore, QtGui

from ftrack_qt.widgets.selectors.version_selector import VersionSelector
from ftrack_qt.widgets.thumbnails import AssetVersionThumbnail


class AssetVersionSelection(QtWidgets.QFrame):
    '''Widget representing an asset, with version selector, within the list,
    for user selection'''

    version_changed = QtCore.Signal(object)
    '''Signal emitted when version is changed, asset_version_id is passed'''

    @property
    def enable_version_select(self):
        '''Return if version_combobox is enabled'''
        return self._version_combobox.isEnabled()

    @enable_version_select.setter
    def enable_version_select(self, value):
        '''Set if version_combobox is enabled'''
        self._version_combobox.setEnabled(value)

    @property
    def version(self):
        '''Return selected version'''
        if self._version_combobox:
            return self._version_combobox.version

    @property
    def asset_id(self):
        '''Return the id of the asset.'''
        return self._asset_id

    def __init__(self, asset_name, asset_id, versions, server_url):
        '''Initialize the AssetVersionSelection widget.'''
        super(AssetVersionSelection, self).__init__()

        self._asset_id = asset_id
        self._asset_name = asset_name
        self._versions = versions
        self._server_url = server_url

        self._thumbnail_widget = None
        self._asset_name_widget = None
        self._version_combobox = None
        self._version_info_widget = None

        self.pre_build()
        self.build()
        self.post_build()

    def pre_build(self):
        '''Set up the layout for the widget.'''
        self.setLayout(QtWidgets.QHBoxLayout())
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(5)

    def build(self):
        '''Build the widget components.'''
        self._thumbnail_widget = AssetVersionThumbnail()
        self._thumbnail_widget.setScaledContents(True)
        self._thumbnail_widget.setMinimumSize(57, 31)
        self._thumbnail_widget.setMaximumSize(57, 31)
        self.layout().addWidget(self._thumbnail_widget)

        self._asset_name_widget = QtWidgets.QLabel(self._asset_name)
        self.layout().addWidget(self._asset_name_widget)

        self._version_combobox = VersionSelector()
        self._version_combobox.set_versions(self._versions)
        self._version_combobox.setMaximumHeight(20)
        self.layout().addWidget(self._version_combobox)

        self._version_info_widget = QtWidgets.QLabel()
        self._version_info_widget.setObjectName('gray')
        self.layout().addWidget(self._version_info_widget, 10)

        self._thumbnail_widget.set_server_url(self._server_url)
        self._thumbnail_widget.load(
            self._version_combobox.version['thumbnail']
        )
        self._version_info_widget.setText(
            f"{self._version_combobox.version['user_first_name']} "
            f"{self._version_combobox.version['user_last_name']} @ "
            f"{self._version_combobox.version['date']}"
        )

    def post_build(self):
        '''Connect signals after building the widget.'''
        self._version_combobox.currentIndexChanged.connect(
            self._on_current_version_changed
        )

    def _on_current_version_changed(self, index):
        '''Handle the event when the current version is changed.'''
        version_dict = self._version_combobox.version
        if version_dict:
            self.version_changed.emit(version_dict)
            self._thumbnail_widget.load(
                self._version_combobox.version['thumbnail']
            )
            self._version_info_widget.setText(
                f"{self._version_combobox.version['user_first_name']} "
                f"{self._version_combobox.version['user_last_name']} @ "
                f"{self._version_combobox.version['date']}"
            )



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

from Qt import QtWidgets, QtCore, QtGui


class TableModel(QtCore.QAbstractTableModel):
    '''Table model for generic data'''

    DATA_ROLE = QtCore.Qt.UserRole + 1

    @property
    def data_items(self):
        '''
        Returns the :obj:`data_items`
        '''
        return self._data_items

    @property
    def headers(self):
        '''
        Returns the :obj:`headers`
        '''
        return self._headers

    def __init__(self, column_mapping, parent=None):
        '''
        Initialise Model.
        *column_mapping*: is a dictionary to map data item keys to header titles
        '''
        super(TableModel, self).__init__(parent=parent)

        self._column_mapping = column_mapping
        self._headers = list(self._column_mapping.keys())
        self._data_items = []
        self._editable_columns = []

    def set_data_items(self, data_items):
        '''
        Reset the model and sets the :obj:`data_items` with the given
        *data_items*
        '''
        self.beginResetModel()
        self._data_items = data_items
        self.endResetModel()

    def set_editable_column(self, column_index):
        '''Sets the given *column_index* as editable'''
        if column_index not in self._editable_columns:
            self._editable_columns.append(column_index)

    def rowCount(self, parent):
        '''Return the row count for the internal data.'''
        if parent.column() > 0:
            return 0

        return len(self.data_items)

    def columnCount(self, parent):
        '''Return the column count for the internal data.'''
        return len(self._headers)

    def data(self, index, role=QtCore.Qt.DisplayRole):
        '''Return the data provided in the given *index* and with *role*'''

        row = index.row()
        column = index.column()

        if not index.isValid():
            return None

        item = self.data_items[row]
        column_name = self._headers[column]

        # style versions
        if role == QtCore.Qt.TextAlignmentRole and column == 0:
            return QtCore.Qt.AlignCenter

        # style the rest
        elif role == QtCore.Qt.DisplayRole:
            if type(self._column_mapping[column_name]) == list:
                # Small function to get the value from recursive keys
                def get_recursive_keys(_item, keys):
                    if len(keys) == 1:
                        return _item[keys[0]]
                    return get_recursive_keys(_item[keys[0]], keys[1:])

                return get_recursive_keys(
                    item, self._column_mapping[column_name]
                )
            return str(item[self._column_mapping[column_name]])

        elif role == QtCore.Qt.EditRole:
            return item

        elif role == self.DATA_ROLE:
            return item

        return None

    def headerData(self, col, orientation, role):
        '''Provide header data'''
        if (
            orientation == QtCore.Qt.Horizontal
            and role == QtCore.Qt.DisplayRole
        ):
            return self._headers[col].capitalize()
        return None

    def flags(self, index):
        '''Set :obj:`self._editable_columns` editable'''
        if index.column() in self._editable_columns:
            return QtCore.Qt.ItemIsEditable | QtCore.Qt.ItemIsEnabled
        else:
            return QtCore.Qt.ItemIsEnabled | QtCore.Qt.ItemIsSelectable



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

import logging

from Qt import QtWidgets, QtCore

from ftrack_qt.widgets.buttons import CircularButton


class ListSelector(QtWidgets.QWidget):
    current_item_changed = QtCore.Signal(object)
    refresh_clicked = QtCore.Signal()

    @property
    def label(self):
        return self._label

    @label.setter
    def label(self, value):
        self._label = str(value)

    @property
    def no_items_label(self):
        return self._no_items_label_widget.text()

    @no_items_label.setter
    def no_items_label(self, value):
        self._no_items_label_widget.set_text = str(value)

    def __init__(self, label, parent=None):
        '''
        Initialize list_selector widget

        :param parent: The parent dialog or frame
        '''
        super(ListSelector, self).__init__(parent=parent)

        self.logger = logging.getLogger(
            __name__ + '.' + self.__class__.__name__
        )
        self._label = label
        self._label_widget = None
        self._combo_box_selector = None
        self._refresh_button = None
        self._no_items_label_widget = None

        self.pre_build()
        self.build()
        self.post_build()

    def pre_build(self):
        main_layout = QtWidgets.QHBoxLayout()
        self.setLayout(main_layout)

    def build(self):
        # Create Label
        self._label_widget = QtWidgets.QLabel(self.label)

        # Create the no items label
        # TODO: finish this
        self._no_items_label_widget = QtWidgets.QLabel()
        self.no_items_label = '<html><i>No items available</i></html>'

        # Create the combo box
        self._combo_box_selector = QtWidgets.QComboBox()
        self._combo_box_selector.setToolTip(self.label)

        # Create the refresh button
        self._refresh_button = CircularButton('sync')

        # Set up the layout
        self.layout().addWidget(self._label_widget)
        self.layout().addWidget(self._combo_box_selector)
        self.layout().addWidget(self._refresh_button)

    def post_build(self):
        self._combo_box_selector.currentIndexChanged.connect(
            self._on_current_index_changed_callback
        )
        self._refresh_button.clicked.connect(self._on_refresh_callback)

    def clear_items(self):
        self._combo_box_selector.clear()

    def add_item(self, item_text):
        if self._combo_box_selector.count() == 0:
            # Add first empty object
            self._combo_box_selector.addItem("-- {} --".format(self.label))
        self._combo_box_selector.addItem(item_text)

    def set_current_item(self, item_text):
        if not item_text:
            self._combo_box_selector.setCurrentIndex(0)
        for index in range(0, self._combo_box_selector.count()):
            if self._combo_box_selector.itemText(index) == item_text:
                self._combo_box_selector.setCurrentIndex(index)
                break

    def set_current_item_index(self, item_index):
        self._combo_box_selector.setCurrentIndex(item_index)

    def current_item_text(self):
        return self._combo_box_selector.currentText()

    def current_item_index(self):
        return self._combo_box_selector.currentIndex()

    def _on_current_index_changed_callback(self, index):
        self.current_item_changed.emit(self._combo_box_selector.currentText())

    def _on_refresh_callback(self):
        self.refresh_clicked.emit()



# :coding: utf-8
# :copyright: Copyright (c) 2024 ftrack

from Qt import QtGui, QtCore, QtWidgets

from ftrack_qt.utils.widget import set_property
from ftrack_qt.widgets.buttons import CircularButton
from ftrack_qt.utils.layout import recursive_clear_layout


class SearchBox(QtWidgets.QFrame):
    '''
    Widget displaying a search box, that can be collapsed and expanded.
    '''

    inputUpdated = QtCore.Signal(object)  # User has update input
    clear = QtCore.Signal()  # Clear button was pressed

    @property
    def text(self):
        '''Retrieve the search text'''
        return self._input.text() if self._input else ''

    @text.setter
    def text(self, value):
        '''Set the search text to *value*'''
        if self._input:
            self._input.setText(value)

    def __init__(self, collapsed=True, collapsable=True, parent=None):
        '''
        Initialize the search widget

        :param collapsed: If True, search box should start collapsed (default)
        :param collapsable: If True, search can be collapsed by user.
        :param parent: The parent dialog or frame
        '''
        super(SearchBox, self).__init__(parent=parent)

        self._collapsed = collapsed
        self._collapsable = collapsable

        self._search_button = None
        self._input = None

        self.pre_build()
        self.build()
        self.post_build()

    def pre_build(self):
        '''Prepare general layout.'''
        self.setLayout(QtWidgets.QHBoxLayout(self))
        self.layout().setContentsMargins(4, 1, 5, 1)
        self.layout().setSpacing(1)
        self.setMaximumHeight(33)
        self.setMinimumHeight(33)

    def build(self):
        '''Build widgets and parent them.'''
        self.rebuild()

    def post_build(self):
        '''Post Build ui method for events connections.'''
        self._search_button.clicked.connect(self._on_search_clicked)

    def rebuild(self):
        '''Remove current widgets, clear input'''
        recursive_clear_layout(self.layout())
        set_property(self, 'collapsed', 'true' if self._collapsed else 'false')
        if self._collapsed:
            # Just the circular search button
            self.layout().addStretch()
            self._input = None
            self._search_button = CircularButton('search')
        else:
            self._search_button = CircularButton('search', diameter=30)

        if self._collapsable:
            self.layout().addWidget(self._search_button)

        if not self._collapsed:
            # A bordered input field filling all space, with input and a clear button
            self._input = QtWidgets.QLineEdit()
            self._input.setReadOnly(False)
            self._input.textChanged.connect(self._on_input_changed)
            self._input.setPlaceholderText("Type to search")
            self._input.setFocus()
            self.layout().addWidget(self._input, 100)
            if not self._collapsable:
                self.layout().addWidget(self._search_button)

    def _on_search_clicked(self):
        '''User clicked search button, update collapsed state'''
        if self._collapsable:
            self._collapsed = not self._collapsed
            self.rebuild()
            self.inputUpdated.emit('')

    def _on_input_changed(self):
        '''Search input text changed'''
        self.inputUpdated.emit(self._input.text())

    def _on_clear_clicked(self):
        '''Clear search input'''
        self._input.setText('')
        self._input.setFocus()
        self.clear.emit()
